#include "stdafx.h"
#include "CImageSegmentationIndexFile.h"
#include <cstdio>
#include <string>

#ifdef WIN32

#include <string.h>

#else

#include <strings.h>
#define stricmp strcasecmp

#endif

using namespace std;

namespace segmentation {

const char CImageSegmentationIndexFile::segFileExt[2][9]={".bsi",".seg"}; // write extentions as lowercase !!!

/*
BSI file format

All numbers are signed 32 bit long integer.
File format:
0x00425349          4 bytes - 0BSI 
file version(1)		4 bytes
image_width         4 bytes
image_height        4 bytes
segment_index_for_each_pixel  4 * image_width * image_height (>=0)
*/

#define BSI_FILE_FLAG 0x00425349
#define CURRENT_FILE_VERSION 1


FILE * CImageSegmentationIndexFile::ReadBSIFileHeader(const char *filename, long & width, long & height)
{
	signed long int l,ver;
  bool not_a_bsi_file = 0;
	
	FILE *f=fopen(filename,"rb");

	if (!f)
		return NULL;

  if (sizeof(signed long int) != 4)
  {
    assert(sizeof(signed int) == 4); //we have to have at least one data type with size = 4 bytes

    int intl = (signed int) l;
    int intver = (signed int) ver;
    int intwidth = (signed int) width;
    int intheight = (signed int) height;

    if (fread(&intl, sizeof(intl), 1, f)!=1)
      not_a_bsi_file = 1;

    if (intl!=(signed int) BSI_FILE_FLAG)
      not_a_bsi_file = 1;

    if (fread(&intver, sizeof(intver), 1, f)!=1)
      not_a_bsi_file = 1;

    if (intver!=1)
      not_a_bsi_file = 1;

    if (fread(&intwidth, sizeof(intwidth), 1, f)!=1)
      not_a_bsi_file = 1;

    if (fread(&intheight, sizeof(intheight), 1, f)!=1)
      not_a_bsi_file = 1;

    l = intl;
    ver = intver;
    width = intwidth;
    height = intheight;
  }
  else
  {
  	if (fread(&l, sizeof(l), 1, f)!=1)
      not_a_bsi_file = 1;

  	if (l!=BSI_FILE_FLAG)
      not_a_bsi_file = 1;

  	if (fread(&ver, sizeof(ver), 1, f)!=1)
      not_a_bsi_file = 1;

    if (ver!=1)
      not_a_bsi_file = 1;

    if (fread(&width, sizeof(width), 1, f)!=1)
      not_a_bsi_file = 1;

  	if (fread(&height, sizeof(height), 1, f)!=1)
      not_a_bsi_file = 1;
  }

  if (not_a_bsi_file == 1) {
   fclose(f);
   return NULL;
  }

	return f;
}

bool CImageSegmentationIndexFile::ReadBSIFileData(FILE *f, long * indexArray, long width, long height)
{
  if (sizeof(signed long int) != 4)
  {
    assert(sizeof(signed int) == 4); //we have to have at least one data type with size = 4 bytes

    signed int * intindexArray = new signed int [width*height];

    if (fread(intindexArray, sizeof(*intindexArray), width*height, f)!=width*height)
      return false;

    for (int i=0; i<width*height; i++)
    {
      indexArray[i] =  intindexArray[i];
    }

    if (intindexArray != NULL)
    {
      delete intindexArray;
      intindexArray = 0;
    }
  }
  else
  {
    if (fread(indexArray, sizeof(*indexArray), width*height, f)!=width*height)
      return false;
  }

  return true;
}


FILE * CImageSegmentationIndexFile::ReadSEGFileHeader(const char *filename, long & width, long & height)
{
	width=-1;
	height=-1;

	FILE *f=fopen(filename,"rb");

	if (!f)
		return NULL;

	char headerCommand[1024], commandParam[1024];

	while (1)
	{
		if (fscanf(f,"%s",headerCommand)==EOF)
			goto not_a_seg_file; // we did not find the "data" command

		fgets(commandParam,1023,f);

		if (stricmp(headerCommand,"format")==0)
		{
			if (strstr(commandParam,"ascii")==NULL || strstr(commandParam,"cr")==NULL)
				goto not_a_seg_file;
		}
		else if (stricmp(headerCommand,"width")==0)
		{
			width=atoi(commandParam);
		}
		else if (stricmp(headerCommand,"height")==0)
		{
			height=atoi(commandParam);
		}
		else if (stricmp(headerCommand,"invert")==0)
		{
			if (strstr(commandParam,"0")==NULL) // we only know this type
				goto not_a_seg_file;
		}
		else if (stricmp(headerCommand,"flipflop")==0)
		{
			if (strstr(commandParam,"0")==NULL) // we only know this type
				goto not_a_seg_file;
		}
		else if (stricmp(headerCommand,"data")==0)
		{
			break; // header done
		}
		//else
			// other commands
			// date, image, user, segments, gray
	}

	if (width<=0 || height<=0)
		goto not_a_seg_file;

	return f;

not_a_seg_file:
	fclose(f);
	return NULL;
}

bool CImageSegmentationIndexFile::ReadSEGFileData(FILE *f, long * indexArray, long width, long height)
{
	char lineFromFile[1024];
	long pixelCount, pixelIndex;

	// mark all pixels as "no segment yet"
	for (pixelCount = width*height, pixelIndex = 0; pixelIndex<pixelCount; pixelIndex++)
		indexArray[pixelIndex]=-1;

	while (fgets(lineFromFile,1023,f))
	{
		long s,r,c1,c2;

		// end of file ?
		if (sscanf(lineFromFile, "%d %d %d %d", &s, &r, &c1, &c2)!=4)
			break;

		// problem in the data
		if (r<0 || r>=height || c1<0 || c1>=width || c2<0 || c2>=width || c1>c2)
			return false;

		for (long c=c1, index=r*width+c1; c<=c2; c++, index++)
		{
			if (indexArray[index]!=-1) // this pixel allready assigned-> problem in the file
				return false;

			indexArray[index] = s;
		}
	}

	// check that all pixels are set
	for (pixelCount = width*height, pixelIndex = 0; pixelIndex<pixelCount; pixelIndex++)
		if (indexArray[pixelIndex]==-1)
			return false;

	return true;
}

FILE * CImageSegmentationIndexFile::ReadFileHeader(FILE_FORMAT fileFormat, const char *filename, long & width, long & height)
{
	switch (fileFormat)
	{
	case bsi_format:
		return ReadBSIFileHeader(filename, width, height);

	case martin_seg_format:
		return ReadSEGFileHeader(filename, width, height);
	}

	return NULL;
}


bool CImageSegmentationIndexFile::ReadFileData(FILE_FORMAT fileFormat, FILE *f, long * indexArray, long width, long height)
{
	switch (fileFormat)
	{
	case bsi_format:
		return ReadBSIFileData(f, indexArray, width, height);

	case martin_seg_format:
		return ReadSEGFileData(f, indexArray, width, height);
	}

	return false;
}


long CImageSegmentationIndexFile::Load(const char *filename, long & width, long & height, long * indexArray)
{
	assert(1); // this function has not been tested+used yet

	// do we recognise the file format?
	FILE_FORMAT fileFormat = WhichSegmentationIndexFile(filename);

	if (fileFormat == unknown_format)
		return -1;

	long w,h;

	FILE * f = ReadFileHeader(fileFormat, filename, w, h);
	// success?
	if (!f)
		return -1;

	// good size?
	if (!indexArray || w!=width || h!=height)
	{
		width=w;
		height=h;

		fclose(f);
		return 0;
	}

	// read the file data
	bool dataReadOk = ReadFileData(fileFormat, f, indexArray, w, h);

	fclose(f);

	return dataReadOk ? 1 : -1;
}

long CImageSegmentationIndexFile::Load(const char *filename, TImageSegmentation<long> &segmentation)
{
	// do we recognise the file format?
	FILE_FORMAT fileFormat = WhichSegmentationIndexFile(filename);

	if (fileFormat == unknown_format)
		return -1;

	long w, h;

	FILE * f = ReadFileHeader(fileFormat, filename, w, h);
	// success?
	if (!f)
		return -1;

	// allocate size for data
	segmentation.Init(w, h);

	// read the file data
	bool dataReadOk = ReadFileData(fileFormat, f, *segmentation, w, h);

	fclose(f);

	return dataReadOk ? 1 : -1;
}

bool CImageSegmentationIndexFile::Save(const char *filename, const long width, const long height, const long * indexArray, FILE_FORMAT format)
{
	assert(format==bsi_format); // we only know this right now
// 	assert(sizeof(signed long int)==4); // we don't know what future compilers will have
	
	size_t bw=0;
	FILE *f=fopen(filename,"wb");
	
	if (!f)
		return false;	

	// convert format to 4 bytes so the image segmentation
	// index file will be cross-platform compatible
	if (sizeof(signed long int) != 4)
	{
		assert(sizeof(signed int) == 4); //we have to have at least one data type with size = 4 bytes

		signed int l;

		l=BSI_FILE_FLAG;
		bw=fwrite(&l, sizeof(l), 1, f);

		l=CURRENT_FILE_VERSION;
		bw+=fwrite(&l, sizeof(l), 1, f);

		signed int intwidth = (signed int) width;
		signed int intheight = (signed int) height;
		signed int * intindexArray = new signed int [width*height];
		for (int i=0; i<width*height; i++)
		{
			intindexArray[i] = (signed int) indexArray[i];
		}

		bw+=fwrite(&intwidth, sizeof(intwidth), 1, f);
		bw+=fwrite(&intheight, sizeof(intheight), 1, f);

		bw+=fwrite(intindexArray, sizeof(*intindexArray), width*height, f);

		if (intindexArray != 0)
		{
			delete intindexArray;
			intindexArray = 0;
		}
	}
	else
	{
		signed long int l;

		l=BSI_FILE_FLAG;
		bw=fwrite(&l, sizeof(l), 1, f);

		l=CURRENT_FILE_VERSION;
		bw+=fwrite(&l, sizeof(l), 1, f);

		bw+=fwrite(&width, sizeof(width), 1, f);
		bw+=fwrite(&height, sizeof(height), 1, f);

		bw+=fwrite(indexArray, sizeof(*indexArray), width*height, f);
	}

	fclose(f);

	// did we write the correct number of items?
	if (bw!=4+width*height)
		return false;

	return true;
}

bool CImageSegmentationIndexFile::Save(const char *filename, const TImageSegmentation<long> &segmentation, FILE_FORMAT format)
{
	return Save(filename,segmentation.Width(),segmentation.Height(),*segmentation, format);
}


CImageSegmentationIndexFile::FILE_FORMAT CImageSegmentationIndexFile::WhichSegmentationIndexFile(const char * filename)
{
	std::string tempS(filename);

	std::transform(tempS.begin(),tempS.end(),tempS.begin(),tolower);

	for (int i=0;i<format_count;i++)
		if (tempS.find(segFileExt[i])!=std::string::npos)
			return (FILE_FORMAT)i;

	return unknown_format;
}


}

