#ifndef __CImageSegmentationIndexFile_h_2005_03_23_15_04_W_AI__
#define __CImageSegmentationIndexFile_h_2005_03_23_15_04_W_AI__

#include "TImageSegmentation.h"

namespace segmentation {

// input/ouput filter for image segmentations (a segment index for each pixel)
class CImageSegmentationIndexFile
{
public:
	// extentions of file read+written
	static const char segFileExt[2][9];

	enum FILE_FORMAT
	{
		unknown_format = -1,
		bsi_format = 0,
		martin_seg_format = 1, // David Martin - Berkley Image Database
		format_count = 2
	};

	// read image segment index falues from specified file
	// - if the indexArray pointer is NULL or the width & height do not match the information
	//   in the file, the values of width and height are corrected and 0 is returned
	// - if some file error appears -1 is returned
	// - if the dimensions match, width*height values are written to indexArray and 1 is returned
	static
	long Load(const char *filename, long & width, long & height, long * indexArray = NULL);


	// reads the segmentation to the supplied segmentation object
	// - modifies oject (reallocates) if necessary
	// - only file error (-1) or 1 are possible return values
	static
	long Load(const char *filename, TImageSegmentation<long> &segmentation);

	// write the image segment index to the specified file
	// - width*height values are written from indexArray
	// - returns true on success
	// - false means file error
	static
	bool Save(const char *filename, const long width, const long height, const long * indexArray, FILE_FORMAT format = bsi_format);

	// write the image segment index to the specified file
	// - width*height values are written from indexArray
	// - returns true on success
	// - false means file error
	static
	bool Save(const char *filename, const TImageSegmentation<long> &segmentation, FILE_FORMAT format = bsi_format);

	// returns the format if the filename specifies a known segmentation file
	// unknown_format otherwise
	static
	FILE_FORMAT WhichSegmentationIndexFile(const char * filename);


protected:
	// reads the file header (file format specified)
	static 
	FILE * ReadFileHeader(FILE_FORMAT fileFormat, const char *filename, long & width, long & height);

	// reads the file data (file format specified)
	static
	bool ReadFileData(FILE_FORMAT fileFormat, FILE *f, long * indexArray, long width, long height);

	// read header of BSI files and get width and height
	static
	FILE * ReadBSIFileHeader(const char *filename, long & width, long & height);

	// read segmentation data from BSI files
	static
	bool ReadBSIFileData(FILE *f, long * indexArray, long width, long height);

	// read header of Martin's SEG files and get width and height
	static
	FILE * ReadSEGFileHeader(const char *filename, long & width, long & height);

	// read segmentation data from Martin's SEG files
	static
	bool ReadSEGFileData(FILE *f, long * indexArray, long width, long height);

};

} // namespace segmentation

#endif

